"""Base constructs for build package objects
"""
import re
import string
import json
from typing import (
    Optional,
    TypeVar,
    Generic,
    Any,
    Union,
    AbstractSet,
    Iterable,
        )
from collections.abc import Mapping, KeysView, ValuesView, ItemsView
from collections import Counter
from pydantic import BaseModel, Field
from pydantic.generics import GenericModel
from bgameb.errors import ComponentNameError, ComponentClassError
from loguru._logger import Logger
from loguru import logger


logger.disable('bgameb')


def log_enable(
    log_path: str = './logs/game.log',
    log_level: str = 'DEBUG'
        ) -> None:
    """Enable logging

    Args:
        log_path (str, optional): path to log file.
                                  Defaults to './logs/game.log'.
        log_level (str, optional): logging level. Defaults to 'DEBUG'.
    """
    logger.remove()
    logger.add(
        sink=log_path,
        level=log_level,
        format='{extra[classname]} -> func {function} | ' +
        '{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}',
    )
    logger.enable('bgameb')


IntStr = Union[int, str]
AbstractSetIntStr = AbstractSet[IntStr]
MappingIntStrAny = Mapping[IntStr, Any]


# TODO: test me
class PropertyBaseModel(BaseModel):
    """
    Serializing properties with pydantic
    https://github.com/samuelcolvin/pydantic/issues/935
    https://github.com/pydantic/pydantic/issues/935#issuecomment-554378904
    https://github.com/pydantic/pydantic/issues/935#issuecomment-1152457432
    """
    @classmethod
    def get_properties(cls):
        return [
            prop for prop
            in dir(cls)
            if isinstance(getattr(cls, prop), property)
            and prop not in ("__values__", "fields")
                ]

    def dict(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> dict[str, Any]:
        attribs = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none
        )
        props = self.get_properties()
        # Include and exclude properties
        if include:
            props = [prop for prop in props if prop in include]
        if exclude:
            props = [prop for prop in props if prop not in exclude]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})

        return attribs


class Base(PropertyBaseModel):
    """Base class for game, players, tools and items

    ..
        Attr:

            id (str): id of stuff

            _counter (Counter): Counter object.
                                Isn't represented in final json or dict.
                                Is initialized automaticaly by __init__.
                                Counter is a collection.Counter.

            _logger (Logger): loguru logger

        Counter is a `collection.Counter
        <https://docs.python.org/3/library/collections.html#collections.Counter>`_
    """
    id: str
    _counter: Counter[Any] = Field(default_factory=Counter)
    _logger: Logger = Field(...)

    def __init__(self, **data):
        super().__init__(**data)

        self._counter = Counter()
        self._logger = logger.bind(classname=self.__class__.__name__)

    class Config:
        underscore_attrs_are_private = True


class BaseGame(Base):
    """Base class for games
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._logger.info('===========NEW GAME============')
        self._logger.info(f'{self.__class__.__name__} created.')


class BasePlayer(Base):
    """Base class for players
    """


class BaseItem(Base):
    """Base class for items (like dices or cards)
    """


V = TypeVar('V', bound=BaseItem)


class Components(GenericModel, Generic[V], Mapping[str, V]):
    """Components mapping represents a collection of objects,
    used for create instance of game items, like dices or decks
    """

    def __init__(
        self,
        *args: tuple[dict[str, V]],
        **kwargs: dict[str, V]
            ) -> None:
        for arg in args:
            if isinstance(arg, dict):
                for k, v, in arg.items():
                    self.__dict__[k] = v
            else:
                raise AttributeError('Args must be a dict of dicts')
        if kwargs:
            for k, v, in kwargs.items():
                self.__dict__[k] = v

    def __iter__(self):
        return iter(self.__dict__)

    def __setattr__(self, attr: str, value: V) -> None:
        self.__setitem__(attr, value)

    def __getattr__(self, attr: str) -> V:
        try:
            return self.__getitem__(attr)
        except KeyError:
            raise AttributeError(attr)

    def __delattr__(self, attr: str) -> None:
        try:
            self.__delitem__(attr)
        except KeyError:
            raise AttributeError(attr)

    def __setitem__(self, attr: str, value: V) -> None:
        if issubclass(value.__class__, BaseItem):
            name = self._make_name(attr)
            if name not in self.__dict__.keys():
                self.__dict__[name] = value
            else:
                raise ComponentNameError(name)
        else:
            raise ComponentClassError(value)

    def __getitem__(self, attr: str) -> V:
        return self.__dict__[attr]  # type: ignore

    def __delitem__(self, attr: str) -> None:
        del self.__dict__[attr]

    def __repr__(self) -> str:
        items = list(
            {f"{k}: {v!r}" for k, v in self.items()}
                )
        return f"{{{', '.join(items)}}}"

    def __len__(self) -> int:
        return len(self.__dict__)

    def keys(self) -> KeysView[str]:
        return self.__dict__.keys()

    def values(self) -> ValuesView[V]:
        return self.__dict__.values()

    def items(self) -> ItemsView[str, V]:
        return self.__dict__.items()

    def to_json(self) -> str:
        return json.dumps(self.__dict__, default=lambda c: c.dict())

    def _is_valid(self, name: str) -> bool:
        """Chek is name of stuff contains correct symbols
        match [a-zA-Z_][a-zA-Z0-9_]*$ expression:

            * a-z, A-Z, 0-9 symbols
            * first letter not a number amd not a _
            * can be used _ symbol in subsequent symbols

        Args:
            name (str): name of stuff

        Raises:
            ComponentNameError: name is not valid

        Returns:
            Trye: is valid
        """
        if not re.match("[a-z][a-z0-9_]*$", str(name)):
            raise ComponentNameError(name)
        return True

    def _make_name(self, name: str) -> str:
        """
        Replace spaces and other specific characters
        in the name with _

        Args:
            name (str): name of stuff

        Returns:
            name (str): safe name of stuff
        """
        name = str(name).lower()
        available = set(string.ascii_letters.lower() + string.digits + '_')

        if " " in name:
            name = name.replace(' ', '_')

        diff = set(name).difference(available)
        if diff:
            for char in diff:
                name = name.replace(char, '_')

        self._is_valid(name)

        return name

    def update(
        self,
        stuff: V,
        name: Optional[str] = None,
            ) -> None:
        """Update Component dict with safe name

        Args:
            stuff (BaseItem): item added to Components
            name (Optional[str]). keys to update dict. Defult to None.
        """
        if name is None:
            name = self._make_name(stuff.id)
        else:
            name = self._make_name(name)

        comp = stuff.__class__(**stuff.dict())
        self.__dict__[name] = comp

    @property
    def ids(self) -> list[str]:
        """Get ids of all items in Components

        Returns:
            list[str]: list of stuff ids
        """
        return [stuff.id for stuff in self.values()]

    def by_id(self, id: str) -> Optional[V]:
        """Get item object by its id

        Args:
            id (str): item id

        Returns:
            V, optional: item object
        """
        for comp in self.values():
            if comp.id == id:
                return comp
        return None


class BaseTool(Base, GenericModel, Generic[V]):
    """Base class for game tools
    """
    current: list[V] = []
    last: Optional[V] = None

    @property
    def current_ids(self) -> list[str]:
        """Get ids of current items

        Returns:
            list[str]: list ids of current
        """
        return [item.id for item in self.current]

    @property
    def last_id(self) -> Optional[str]:
        """Get id of last

        Returns:
            Optional[str]: id
        """
        if self.last is not None:
            return self.last.id
        return None

    def _item_replace(self, item: V) -> V:
        """Get item replaced copy

        Returns:
            Item (BaseItem): an item object
        """
        return item.__class__(**item.dict())

    def by_id(self, id: str) -> list[V]:
        """Get item from current by its id

        Args:
            id (str): item id

        Returns:
            list[BaseItem]: items
        """
        return [item for item in self.current if item.id == id]

    def clear(self) -> None:
        """Clear the current and last
        """
        self.current.clear()
        self.last = None
        self._logger.debug('Current and last clear!')

    def count(self, item_id: str) -> int:
        """Count the number of current items with given id.

        Args:
            item_id (str): an item id

        Returns:
            int: count of items
        """
        count = self.current_ids.count(item_id)
        self._logger.debug(f'Count of {item_id} in current is {count}')
        return count

    def pop(self) -> V:
        """Remove and return an item from the current.
        If no items are present, raises an IndexError.

        Returns:
            BaseItem: an item object
        """
        self.last = self.current.pop()
        self._logger.debug(f'{self.last.id} is poped from current')
        return self.last


class BaseToolExtended(BaseTool[V], Generic[V]):
    """Extends BaseTool class by list-like methods
    """

    def append(self, item: V) -> None:
        """Append item to current

        Args:
            item (BaseItem): appended items
        """
        item = self._item_replace(item)
        self.current.append(item)  # type: ignore
        self._logger.debug(f'To current is appended item: {item.id}')

    def extend(self, items: Iterable[V]) -> None:
        """Extend the current by appending items
        started from the left side of iterable.

        Args:
            items (Iterable[BaseItem]): iterable with items
        """
        items = [self._item_replace(item) for item in items]
        self.current.extend(items)  # type: ignore
        self._logger.debug(
            f'Current are extended by {[item.id for item in items]} from right'
                )

    def index(
        self,
        item_id: str,
        start: int = 0,
        end: Optional[int] = None
            ) -> int:
        """Return the position of item in the current
        (after index start and before index stop).
        Returns the first match or raises ValueError if not found.

        Args:
            item_id (str): an item id
            start (int): start index. Default to 0.
            end (int, optional): stop index. Default to None.

        Returns:
            int: index of the the first match
        """
        names = self.current_ids
        ind = names.index(item_id, start) if end is None \
            else names.index(item_id, start, end)
        self._logger.debug(f'Index of {item_id} in current is {ind}')
        return ind

    def insert(self, item: V, pos: int) -> None:
        """Insert item into the current at given position.

        Args:
            item (BaseItem): an item object
            pos (int): position
        """
        item = self._item_replace(item)
        self.current.insert(pos, item)  # type: ignore
        self._logger.debug(f'To current is inserted {item.id} on {pos=}')

    def remove(self, item_id: str) -> None:
        """Remove the first occurrence of item from current.
        If not found, raises a ValueError.

        Args:
            item_id (str): an item id
        """
        ind = self.index(item_id)
        item = self.current[ind]
        self.current.remove(item)
        self._logger.debug(f'Is removed from current {item_id}')

    def reverse(self) -> None:
        """Reverse the items in the current.
        """
        self.current.reverse()
        self._logger.debug('Current is reversed')



"""Game tools classes
"""
import random
from pydantic import Field, PositiveInt
from collections import deque
from collections.abc import KeysView
from heapq import heappop, heappush
from typing import Optional, Iterable, Union, Any
from bgameb.base import BaseTool, BaseToolExtended, Components
from bgameb.items import Card, Dice, Step
from bgameb.errors import ArrangeIndexError


class Shaker(BaseToolExtended[Dice]):
    """Shaker object

    ..
        Attr:

            current (list[Dice]): Current dices representation of shaker.
                                  This making from Components items.

            last (Dice), optional: last dice removed from current.

            last_roll (dict[str, list[PositiveInt]), optional:
                last roll result.

            last_roll_mapped (dict[str, list[Any]]), optional:
                last mapped roll result.
    """
    last_roll: dict[str, list[PositiveInt]] = {}
    last_roll_mapped: dict[str, list[Any]] = {}

    def deal(
        self,
        components: Components[Dice],
        items: Optional[list[str]] = None
            ) -> 'Shaker':
        """Deal new shaker current. The current is cleared
        before deal.

        Args:
            components (Components): game components
            items (Optional[list[str]]): item ids

        Returns:
            Shaker
        """
        self.clear()

        if not items:
            for stuff in components.values():
                if issubclass(stuff.__class__, Dice):
                    self.append(stuff)
        else:
            for id in items:
                if id in components.ids:
                    comp = components.by_id(id)
                    if issubclass(comp.__class__, Dice):
                        self.append(comp)

        self._logger.debug(f'Is deal current: {self.current_ids}')
        return self

    def roll(self) -> dict[str, list[int]]:
        """Roll all stuff in shaker and return results

        Return:
            dict[str, list[int]]: result of roll

        .. code-block::
            :caption: Example:

                {
                    "six_dice": [5, 3, 2, 5],
                    "twenty_dice": [2, 12, 4],
                }
        """
        self.last_roll = {}

        for item in self.current:
            self.last_roll[item.id] = item.roll()  # type: ignore

        self._logger.debug(f'Result of roll: {self.last_roll}')

        return self.last_roll

    def roll_mapped(self) -> dict[str, list[Any]]:
        """Roll all stuff in shaker and return mapped results.
        If any stuff unmaped - empty list returned for this item.

        Returns:
            dict[str, list[Any]]: result of roll
        """
        self.last_roll_mapped = {}

        for item in self.current:
            self.last_roll_mapped[item.id] = item.roll_mapped()  # type: ignore

        self._logger.debug(f'Result of roll: {self.last_roll_mapped}')

        return self.last_roll_mapped


class Deck(BaseToolExtended[Card]):
    """Deck object

    ..
        You can add cards, define it counts and deal a deck.
        Result is saved in current attr as deque object. This object
        has all methods of
        `python deque
        <https://docs.python.org/3/library/collections.html#deque-objects>`_

        Attr:

            current (Deque[Card]): Current cards representation of deck.
                                   This making from Component items.

            last (Card), optional: last card, removed from current.
    """
    current: deque[Card] = Field(default_factory=deque)  # type: ignore

    def _item_replace(self, item: Card) -> Card:
        """Get replaced copy of card

        Args:
            item (Card): a card object

        Returns:
            Card
        """
        item = super()._item_replace(item)
        item.count = 1
        return item

    def deal(
        self,
        components: Components[Card],
        items: Optional[list[str]] = None
            ) -> 'Deck':
        """Deal new deck current. Curent is cleared
        before deal.

        Args:
            components (Components): game components
            items (Optional[list[str]]): list of cards ids

        Returns:
            Deck
        """
        self.clear()

        if not items:
            for stuff in components.values():
                if issubclass(stuff.__class__, Card):
                    for _ in range(stuff.count):
                        self.append(stuff)
        else:
            for id in items:
                if id in components.ids:
                    comp = components.by_id(id)
                    if issubclass(comp.__class__, Card):
                        self.append(comp)

        self._logger.debug(f'Is deal current: {self.current_ids}')
        return self

    def shuffle(self) -> 'Deck':
        """Random shuffle current deck.

        Returns:
            Deck
        """
        random.shuffle(self.current)
        self._logger.debug(f'Is shuffled: {self.current_ids}')
        return self

    def appendleft(self, item: Card) -> None:
        """Add card to the left side of the current deck.

        Args:
            item (Card): a card object
        """
        item = self._item_replace(item)
        self.current.appendleft(item)
        self._logger.debug(f'To left of current is appended card: {item.id}')

    def extendleft(self, items: Iterable[Card]) -> None:
        """Extend the left side of the current deck by appending
        cards started from the right side of iterable.
        The series of left appends results in reversing the order
        of cards in the iterable argument.

        Args:
            items (Iterable[Card]): iterable with cards
        """
        items = [self._item_replace(item) for item in items]
        self.current.extendleft(items)
        self._logger.debug(
            f'Current are extended by {[item.id for item in items]} from left'
                )

    def popleft(self) -> Card:
        """Remove and return a card from the left side of the current deck.
        If no cards are present, raises an IndexError.

        Returns:
            Card
        """
        self.last = self.current.popleft()  # type: ignore
        self._logger.debug(
            f'{self.last.id if self.last else None} '
            'is poped from left of current'
                )
        return self.last  # type: ignore

    def rotate(self, n: int) -> None:
        """Rotate the current deck n steps to the right.
        If n is negative, rotate to the left.

        Args:
            n (int): steps to rotation
        """
        self.current.rotate(n)
        self._logger.debug(f'Current is rotate by {n}')

    def _check_order_len(self, len_: int) -> None:
        """Check is order len valid

        Args:
            l (int): len of order of cards

        Raises:
            ArrangeIndexError: Is given empty order
            ArrangeIndexError: The len of current deque not match order len
        """
        if not len_:
            raise ArrangeIndexError(
                'Given empty order',
                logger=self._logger
                    )

        if len_ > len(self.current):
            raise ArrangeIndexError(
                f'The len of current deque is {len(self.current)} '
                f'but given order has len {len_}.',
                logger=self._logger
                    )

    def _check_is_to_arrange_valid(
        self,
        order: list[str],
        to_arrange: Union[list[str], KeysView[str]]
            ) -> None:
        """Chek is order and deque contains same elements

        Args:
            order (list[str]): ordered list of cards ids
            to_arrange (list[str]): list of deque ids

        Raises:
            ArrangeIndexError: Given card ids and deque ids not match
        """
        if set(to_arrange) ^ set(order):
            raise ArrangeIndexError(
                'Given card ids and deque ids not match.',
                logger=self._logger
                    )

    def reorder(
        self,
        order: list[str],
            ) -> 'Deck':
        """Reorder current deque from right side.

        Args:
            order (list[str]): ordered list of cards ids
            ordered from left side to right

        Returns:
            Deck
        """
        len_ = len(order)
        self._check_order_len(len_)

        to_arrange = {
            self.current[len_-ind-1].id: self.current[len_-ind-1]
            for ind in range(len_)
                }
        self._check_is_to_arrange_valid(order, to_arrange.keys())

        for _ in range(len_):
            self.pop()

        for card in order:
            self.append(to_arrange[card])

        self._logger.debug(f'Is reordered right side of deque: {order}')

        return self

    def reorderleft(
        self,
        order: list[str],
            ) -> 'Deck':
        """Reorder current deque from left side.

        Args:
            order (list[str]): ordered list of cards ids
            ordered from left side to right

        Returns:
            Deck
        """
        len_ = len(order)
        self._check_order_len(len_)

        to_arrange = {
            self.current[ind].id: self.current[ind]
            for ind in range(len_)
                }
        self._check_is_to_arrange_valid(order, to_arrange.keys())

        for _ in range(len_):
            self.popleft()

        for card in reversed(order):
            self.appendleft(to_arrange[card])

        self._logger.debug(f'Is reordered left side of deque: {order}')

        return self

    def reorderfrom(
        self,
        order: list[str],
        start: int,
            ) -> 'Deck':
        """Reorder current deque from right side started
        with given position.

        Args:
            start (int): start of reordering
            order (list[str]): ordered list of cards ids
            ordered from right side to left

        Returns:
            Deck
        """
        len_ = len(order)
        self._check_order_len(len_)
        if start <= 0 or start > len(self.current)-len_:
            raise ArrangeIndexError(
                'Given range is out of current index.',
                logger=self._logger
                    )

        old_deck = list(self.current)
        to_arrange = {
            card.id: card
            for card in old_deck[start:start+len_]
                }
        self._check_is_to_arrange_valid(order, to_arrange.keys())

        for ind1, ind2 in enumerate(range(start, start+len_)):
            self.current[ind2] = to_arrange[order[ind1]]

        return self

    def search(
        self,
        query: dict[str, int],
        remove: bool = True
            ) -> list[Card]:
        """Search for cards in current by its id.

        Args:
            query (dict[str, int]): dict with id of searched
                                    cards and count of searching
            remove (bool): if True - remove searched cards from
                           current deck. Default to True.

        Return:
            List[Card]: list of find cards, equal searching count

        .. code-block::
            :caption: Example:

                game.deck1.search(
                    {'card1': 2,
                     'card2': 1 },
                    remove=False
                    )
        """
        for_deque: deque[Card] = deque()
        result = []

        while True:
            try:
                card = self.current.popleft()
                if card.id in query.keys() and query[card.id] > 0:
                    result.append(card)
                    query[card.id] -= 1
                    if not remove:
                        for_deque.append(card)
                else:
                    for_deque.append(card)
            except IndexError:
                break
        self.current = for_deque
        self._logger.debug(f'Search result: {result}')

        return result

    def get_random(
        self,
        count: int = 1,
        remove: bool = True
            ) -> list[Card]:
        """Get random cards from current deck

        Args:
            count (int, optional): count of random cards. Defaults to 1.
            remove (bool, optional): if True - remove random cards from
                                     current deck. Default to True.

        Returns:
            list[Card]: list of random cards
        """
        if not self.current:
            self._logger.debug(
                'Is empty current deck. Random cards not choosed.'
                    )
            return []
        if not remove:
            result = random.choices(self.current, k=count)
            self._logger.debug(
                f'Random choised cards without remove: {result}'
                    )
            return result
        else:
            result = []
            for _ in range(count):
                if self.current:
                    choice = random.choice(self.current)
                    result.append(choice)
                    self.current.remove(choice)
                else:
                    break
            self._logger.debug(
                f'Random choised cards with remove: {result}'
                    )
            return result


class Steps(BaseTool[Step]):
    """Game steps order object

    ..
        Attr:

            current (list[Step]]):
                Current representation of order in steps.
                This making from Component items.

            last (Step), optional: last poped from current step.
    """
    current: list[Step] = []
    last: Optional[Step] = None

    def deal(
        self,
        components: Components[Step],
        items: Optional[list[str]] = None
            ) -> 'Steps':
        """Clear current order and create new current order

        Args:
            components (Components): game components
            items (Optional[list[str]]): list of stuff ids

        Returns:
            Steps
        """
        self.clear()

        if not items:
            for stuff in components.values():
                if issubclass(stuff.__class__, Step):
                    self.push(stuff)
        else:
            for id in items:
                if id in components.ids:
                    comp = components.by_id(id)
                    if issubclass(comp.__class__, Step):
                        self.push(comp)

        self._logger.debug(f'Is deal current: {self.current_ids}')
        return self

    def push(self, item: Step) -> None:
        """Push Step object to current

        Args:
            item (Step): Step class instance
        """
        replaced = self._item_replace(item)
        heappush(self.current, replaced)

    def pops(self) -> Step:
        """Pop Step object from current with smallest priority

        Returns:
            Step
        """
        self.last = heappop(self.current)
        self._logger.debug(f'{self.last.id} is poped from current')
        return self.last



"""Game dices, coins, cards and other items
"""
import random
from typing import Optional, NoReturn, Any, cast
from pydantic import PositiveInt, NonNegativeInt, ConstrainedInt
from bgameb.base import BaseItem
from bgameb.errors import StuffDefineError


class Step(BaseItem):
    """Game steps or turns

    ..
        Attr:

            priority (NonNegativeInt): priority queue number. Default to 0.
    """
    priority: NonNegativeInt = 0

    def __eq__(self, other: 'Step') -> bool:  # type: ignore[override]
        return self.priority == other.priority

    def __lt__(self, other: 'Step') -> bool:
        return self.priority < other.priority

    def __le__(self, other: 'Step') -> bool:
        return self.priority <= other.priority

    def __ne__(self, other: 'Step') -> bool:  # type: ignore[override]
        return self.priority != other.priority

    def __gt__(self, other: 'Step') -> bool:
        return self.priority > other.priority

    def __ge__(self, other: 'Step') -> bool:
        return self.priority >= other.priority


class Sides(ConstrainedInt):
    """Int subtipe to define sides of dices

    ..
        Attr:

            gt (int): greate than 1 constraint.
    """
    gt = 1


class Dice(BaseItem):
    """Rolling or tossed objects, like dices or coins.

    .. code-block::
        :caption: Example:

            dice = Dice(id='coin', sides=2, mapping={1: 'this', 2: 'that'})

    ..
        Attr:

            count (PositiveInt): count of dices. Default to 1.

            sides (Sides): sides of dice or coin. Default to 2.

            mapping (dict[PositiveInt, Any]):
                optional mapping of roll result. Mapping must define
                values for each side.

            last roll (list[PositiveInt]), optional:
                last roll values.

            last roll mapped (list[Any]), optional:
                last mapped roll values.

            _range (list[PositiveInt]): range of roll, started from 1.

        Raises:

            StuffDefineError: mapping keys is not equal of roll range.
    """
    count: PositiveInt = 1
    sides: Sides = cast(Sides, 2)
    mapping: dict[PositiveInt, Any] = {}
    last_roll: list[PositiveInt] = []
    last_roll_mapped: list[Any] = []
    _range: list[PositiveInt] = []

    def __init__(self, **data):
        super().__init__(**data)
        self._range = list(range(1, self.sides + 1))

        if self.mapping and set(self.mapping.keys()) ^ set(self._range):
            raise StuffDefineError(
                message='Mapping must define values for each side.',
                logger=self._logger
                    )

    def __eq__(self, other: 'Dice') -> bool:  # type: ignore[override]
        return self.sides == other.sides

    def __lt__(self, other: 'Dice') -> bool:
        return self.sides < other.sides

    def __le__(self, other: 'Dice') -> bool:
        return self.sides <= other.sides

    def __ne__(self, other: 'Dice') -> bool:  # type: ignore[override]
        return self.sides != other.sides

    def __gt__(self, other: 'Dice') -> bool:
        return self.sides > other.sides

    def __ge__(self, other: 'Dice') -> bool:
        return self.sides >= other.sides

    def roll(self) -> list[PositiveInt]:
        """Roll and return result

        Returns:
            list[PositiveInt]: result of roll
        """
        self.last_roll = [
            random.choices(self._range, k=1)[0] for _
            in list(range(self.count))
                ]
        return self.last_roll

    def roll_mapped(self) -> list[Any]:
        """Roll and return mapped result

        Returns:
            list[Any]: result of roll
        """
        self.last_roll_mapped = [
            self.mapping[roll] for roll in self.roll()
            if self.mapping.get(roll)
                ]
        return self.last_roll_mapped


class Card(BaseItem):
    """Card objects

    ..
        Attr:

            count (PositiveInt): count of cards. Default to 1.

            is_revealed (bool): is card oppened. Default to False.

            is_active (bool): is card tapped. Default to False.

            side (str, optional): the side of tap. Default to None.

    .. code-block::
        :caption: Example:

            card = Card(id='unique_card')
            card.tap(side='left')
    """
    count: PositiveInt = 1
    is_revealed: bool = False
    is_active: bool = True
    side: Optional[str] = None

    def flip(self) -> 'Card':
        """Face up or face down the card regardles of it condition

        Returns:
            Card
        """
        if self.is_revealed:
            self.is_revealed = False
            self._logger.debug('Card face down.')
        else:
            self.is_revealed = True
            self._logger.debug('Card face up.')
        return self

    def open(self) -> 'Card':
        """Face up the card

        Returns:
            Card
        """
        self.is_revealed = True
        self._logger.debug('Card face up.')
        return self

    def hide(self) -> 'Card':
        """Face down the card

        Returns:
            Card
        """
        self.is_revealed = False
        self._logger.debug('Card face down.')
        return self

    def tap(self, side='right') -> 'Card':
        """Tap the card to the given side

        Args:
            side (str, optional): side to tap. Defaults to 'right'.

        Returns:
            Card
        """
        self.is_active = False
        self.side = side
        self._logger.debug(f'Card taped to side {side}.')
        return self

    def untap(self) -> 'Card':
        """Untap the card

        Returns:
            Card
        """
        self.is_active = True
        self.side = None
        self._logger.debug('Card untaped. Side set to None.')
        return self

    def alter(self) -> NoReturn:
        """Many cards have alter views. For example
        card can have main view, that apply most time of the game
        and second view, that apply only if card played as
        that alternative. For ease of understanding, consider that
        different views of the same card are not related directly
        to each other.
        """
        raise NotImplementedError



"""Custom error classes
"""
from loguru._logger import Logger


class CustomRuntimeError(RuntimeError):
    """Base class for other runtime exceptions
    """


class ComponentNameError(CustomRuntimeError):
    """Given name of component is wrong or name isn't unique.
    """
    def __init__(self, name: str) -> None:
        self.message = f'Stuff with {name=} is exist in ' + \
                        'Component class instance or wrong name of stuff.'
        super().__init__(self.message)


class ComponentClassError(CustomRuntimeError):
    """Given class can't be a part of components.
    """
    def __init__(self, obj_) -> None:
        self.message = f'Given: {obj_} cant be used as part of Components.'
        super().__init__(self.message)


class StuffDefineError(AttributeError):
    """Bad definition of item.
    """
    def __init__(self, message: str, logger: Logger) -> None:
        self.message = message
        logger.exception(self.message)
        super().__init__(self.message)


class ArrangeIndexError(IndexError):
    """Index error for arrange tool.
    """
    def __init__(self, message: str, logger: Logger) -> None:
        self.message = message
        logger.exception(self.message)
        super().__init__(self.message)



"""
Provides bgameb version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update bgameb` to change this file.

from incremental import Version

__version__ = Version("bgameb", 2, 0, 0)
__all__ = ["__version__"]



from bgameb.items import Dice, Card, Step
from bgameb.tools import Shaker, Deck, Steps
from bgameb.players import Player
from bgameb.game import Game
from bgameb.base import log_enable, Components
from bgameb._version import __version__


PUB_VERSION = __version__.public()
